@@ -556,8 +556,7 @@ int network_handshake_setup(struct network *network, struct scan_bss *bss,
 		handshake_state_set_protocol_version(hs, eapol_proto_version);
 	}
 
-	hs->force_default_ecc_group = network->force_default_ecc_group ||
-						bss->force_default_sae_group;
+	hs->force_default_ecc_group = network->force_default_ecc_group;
 
 	/*
 	 * The randomization options in the provisioning file are dependent on
